@@ -19,6 +19,7 @@ from six.moves import urllib
 
 from keystoneauth1 import _utils
 from keystoneauth1 import access
+from keystoneauth1 import adapter
 from keystoneauth1 import discover
 from keystoneauth1 import exceptions
 from keystoneauth1 import fixture
@@ -329,7 +330,7 @@ class CommonIdentityTests(object):
         sb = session.Session()
         discovery_cache = {}
 
-        expected_url = urllib.parse.urljoin(self.TEST_COMPUTE_ADMIN, '/v2.0')
+        expected_url = self.TEST_COMPUTE_ADMIN + '/v2.0'
         for sess in (sa, sb):
 
             disc = discover.get_discovery(
@@ -340,6 +341,31 @@ class CommonIdentityTests(object):
 
         self.assertIn(self.TEST_COMPUTE_ADMIN, discovery_cache.keys())
 
+    def test_discovery_trailing_slash(self):
+        # The discovery cache should treat root urls the same whether they have
+        # a slash or not. If the url is called a second time (meaning the cache
+        # didn't work, we'll hit the 500 error.
+        self.requests_mock.get(
+            'https://example.com', [
+                {'json': self.TEST_DISCOVERY},
+                {'status_code': 500}
+            ])
+
+        sess = session.Session()
+        discovery_cache = {}
+
+        expected_url = 'https://example.com/v2.0'
+
+        for test_endpoint in ('https://example.com', 'https://example.com/'):
+            disc = discover.get_discovery(
+                sess, test_endpoint, cache=discovery_cache)
+            url = disc.url_for(('2', '0'))
+
+            self.assertEqual(expected_url, url)
+
+        self.assertIn('https://example.com', discovery_cache.keys())
+        self.assertNotIn('https://example.com/', discovery_cache.keys())
+
     def test_discovering_with_no_data(self):
         # which returns discovery information pointing to TEST_URL but there is
         # no data there.
@@ -579,6 +605,434 @@ class CommonIdentityTests(object):
         # We should have gotten the version from the URL
         self.assertEqual((3, 0), data.api_version)
 
+    def test_get_all_version_data_all_interfaces(self):
+
+        for interface in ('public', 'internal', 'admin'):
+            # The version discovery dict will not have a project_id
+            disc = fixture.DiscoveryList(v2=False, v3=False)
+            disc.add_nova_microversion(
+                href=getattr(self.TEST_VOLUME.versions['v3'].discovery,
+                             interface),
+                id='v3.0', status='CURRENT',
+                min_version='3.0', version='3.20')
+
+            # Adding a v2 version to a service named volumev3 is not
+            # an error. The service itself is cinder and has more than
+            # one major version.
+            disc.add_nova_microversion(
+                href=getattr(self.TEST_VOLUME.versions['v2'].discovery,
+                             interface),
+                id='v2.0', status='SUPPORTED')
+
+            self.stub_url(
+                'GET', [],
+                base_url=getattr(self.TEST_VOLUME.unversioned,
+                                 interface) + '/',
+                json=disc)
+
+        for url in (
+                self.TEST_COMPUTE_PUBLIC,
+                self.TEST_COMPUTE_INTERNAL,
+                self.TEST_COMPUTE_ADMIN):
+
+            disc = fixture.DiscoveryList(v2=False, v3=False)
+            disc.add_microversion(
+                href=url, id='v2')
+            disc.add_microversion(
+                href=url, id='v2.1',
+                min_version='2.1', max_version='2.35')
+
+            self.stub_url('GET', [], base_url=url, json=disc)
+
+        a = self.create_auth_plugin()
+        s = session.Session(auth=a)
+
+        identity_endpoint = 'http://127.0.0.1:35357/{}/'.format(self.version)
+        data = s.get_all_version_data(interface=None)
+        self.assertEqual({
+            'RegionOne': {
+                'admin': {
+                    'block-storage': [{
+                        'collection': None,
+                        'max_microversion': None,
+                        'min_microversion': None,
+                        'next_min_version': None,
+                        'not_before': None,
+                        'raw_status': 'SUPPORTED',
+                        'status': 'SUPPORTED',
+                        'url': 'https://block-storage.example.com/admin/v2',
+                        'version': '2.0'
+                    }, {
+                        'collection': None,
+                        'max_microversion': '3.20',
+                        'min_microversion': '3.0',
+                        'next_min_version': None,
+                        'not_before': None,
+                        'raw_status': 'CURRENT',
+                        'status': 'CURRENT',
+                        'url': 'https://block-storage.example.com/admin/v3',
+                        'version': '3.0'
+                    }],
+                    'compute': [{
+                        'collection': None,
+                        'max_microversion': None,
+                        'min_microversion': None,
+                        'next_min_version': None,
+                        'not_before': None,
+                        'raw_status': 'stable',
+                        'status': 'CURRENT',
+                        'url': 'https://compute.example.com/nova/admin',
+                        'version': '2.0'
+                    }, {
+                        'collection': None,
+                        'max_microversion': '2.35',
+                        'min_microversion': '2.1',
+                        'next_min_version': None,
+                        'not_before': None,
+                        'raw_status': 'stable',
+                        'status': 'CURRENT',
+                        'url': 'https://compute.example.com/nova/admin',
+                        'version': '2.1'}],
+                    'identity': [{
+                        'collection': None,
+                        'max_microversion': None,
+                        'min_microversion': None,
+                        'next_min_version': None,
+                        'not_before': None,
+                        'raw_status': None,
+                        'status': 'CURRENT',
+                        'url': identity_endpoint,
+                        'version': self.discovery_version,
+                    }]
+                },
+                'internal': {
+                    'baremetal': [{
+                        'collection': None,
+                        'max_microversion': None,
+                        'min_microversion': None,
+                        'next_min_version': None,
+                        'not_before': None,
+                        'raw_status': None,
+                        'status': 'CURRENT',
+                        'url': 'https://baremetal.example.com/internal/',
+                        'version': None
+                    }],
+                    'block-storage': [{
+                        'collection': None,
+                        'max_microversion': None,
+                        'min_microversion': None,
+                        'next_min_version': None,
+                        'not_before': None,
+                        'raw_status': 'SUPPORTED',
+                        'status': 'SUPPORTED',
+                        'url': 'https://block-storage.example.com/internal/v2',
+                        'version': '2.0'
+                    }, {
+                        'collection': None,
+                        'max_microversion': '3.20',
+                        'min_microversion': '3.0',
+                        'next_min_version': None,
+                        'not_before': None,
+                        'raw_status': 'CURRENT',
+                        'status': 'CURRENT',
+                        'url': 'https://block-storage.example.com/internal/v3',
+                        'version': '3.0'
+                    }],
+                    'compute': [{
+                        'collection': None,
+                        'max_microversion': None,
+                        'min_microversion': None,
+                        'next_min_version': None,
+                        'not_before': None,
+                        'raw_status': 'stable',
+                        'status': 'CURRENT',
+                        'url': 'https://compute.example.com/nova/internal',
+                        'version': '2.0'
+                    }, {
+                        'collection': None,
+                        'max_microversion': '2.35',
+                        'min_microversion': '2.1',
+                        'next_min_version': None,
+                        'not_before': None,
+                        'raw_status': 'stable',
+                        'status': 'CURRENT',
+                        'url': 'https://compute.example.com/nova/internal',
+                        'version': '2.1'
+                    }]
+                },
+                'public': {
+                    'block-storage': [{
+                        'collection': None,
+                        'max_microversion': None,
+                        'min_microversion': None,
+                        'next_min_version': None,
+                        'not_before': None,
+                        'raw_status': 'SUPPORTED',
+                        'status': 'SUPPORTED',
+                        'url': 'https://block-storage.example.com/public/v2',
+                        'version': '2.0'
+                    }, {
+                        'collection': None,
+                        'max_microversion': '3.20',
+                        'min_microversion': '3.0',
+                        'next_min_version': None,
+                        'not_before': None,
+                        'raw_status': 'CURRENT',
+                        'status': 'CURRENT',
+                        'url': 'https://block-storage.example.com/public/v3',
+                        'version': '3.0'
+                    }],
+                    'compute': [{
+                        'collection': None,
+                        'max_microversion': None,
+                        'min_microversion': None,
+                        'next_min_version': None,
+                        'not_before': None,
+                        'raw_status': 'stable',
+                        'status': 'CURRENT',
+                        'url': 'https://compute.example.com/nova/public',
+                        'version': '2.0'
+                    }, {
+                        'collection': None,
+                        'max_microversion': '2.35',
+                        'min_microversion': '2.1',
+                        'next_min_version': None,
+                        'not_before': None,
+                        'raw_status': 'stable',
+                        'status': 'CURRENT',
+                        'url': 'https://compute.example.com/nova/public',
+                        'version': '2.1',
+                    }]
+                }
+            }
+        }, data)
+
+    def test_get_all_version_data(self):
+
+        cinder_disc = fixture.DiscoveryList(v2=False, v3=False)
+
+        # The version discovery dict will not have a project_id
+        cinder_disc.add_nova_microversion(
+            href=self.TEST_VOLUME.versions['v3'].discovery.public,
+            id='v3.0', status='CURRENT',
+            min_version='3.0', version='3.20')
+
+        # Adding a v2 version to a service named volumev3 is not
+        # an error. The service itself is cinder and has more than
+        # one major version.
+        cinder_disc.add_nova_microversion(
+            href=self.TEST_VOLUME.versions['v2'].discovery.public,
+            id='v2.0', status='SUPPORTED')
+
+        self.stub_url(
+            'GET', [],
+            base_url=self.TEST_VOLUME.unversioned.public + '/',
+            json=cinder_disc)
+
+        nova_disc = fixture.DiscoveryList(v2=False, v3=False)
+        nova_disc.add_microversion(
+            href=self.TEST_COMPUTE_PUBLIC, id='v2')
+        nova_disc.add_microversion(
+            href=self.TEST_COMPUTE_PUBLIC, id='v2.1',
+            min_version='2.1', max_version='2.35')
+
+        self.stub_url(
+            'GET', [], base_url=self.TEST_COMPUTE_PUBLIC, json=nova_disc)
+        a = self.create_auth_plugin()
+        s = session.Session(auth=a)
+
+        data = s.get_all_version_data(interface='public')
+        self.assertEqual({
+            'RegionOne': {
+                'public': {
+                    'block-storage': [{
+                        'collection': None,
+                        'max_microversion': None,
+                        'min_microversion': None,
+                        'next_min_version': None,
+                        'not_before': None,
+                        'raw_status': 'SUPPORTED',
+                        'status': 'SUPPORTED',
+                        'url': 'https://block-storage.example.com/public/v2',
+                        'version': '2.0'
+                    }, {
+                        'collection': None,
+                        'max_microversion': '3.20',
+                        'min_microversion': '3.0',
+                        'next_min_version': None,
+                        'not_before': None,
+                        'raw_status': 'CURRENT',
+                        'status': 'CURRENT',
+                        'url': 'https://block-storage.example.com/public/v3',
+                        'version': '3.0'
+                    }],
+                    'compute': [{
+                        'collection': None,
+                        'max_microversion': None,
+                        'min_microversion': None,
+                        'next_min_version': None,
+                        'not_before': None,
+                        'raw_status': 'stable',
+                        'status': 'CURRENT',
+                        'url': 'https://compute.example.com/nova/public',
+                        'version': '2.0'
+                    }, {
+                        'collection': None,
+                        'max_microversion': '2.35',
+                        'min_microversion': '2.1',
+                        'next_min_version': None,
+                        'not_before': None,
+                        'raw_status': 'stable',
+                        'status': 'CURRENT',
+                        'url': 'https://compute.example.com/nova/public',
+                        'version': '2.1'
+                    }],
+                }
+            }
+        }, data)
+
+    def test_get_all_version_data_by_service_type(self):
+        nova_disc = fixture.DiscoveryList(v2=False, v3=False)
+        nova_disc.add_microversion(
+            href=self.TEST_COMPUTE_PUBLIC, id='v2')
+        nova_disc.add_microversion(
+            href=self.TEST_COMPUTE_PUBLIC, id='v2.1',
+            min_version='2.1', max_version='2.35')
+
+        self.stub_url(
+            'GET', [], base_url=self.TEST_COMPUTE_PUBLIC, json=nova_disc)
+        a = self.create_auth_plugin()
+        s = session.Session(auth=a)
+
+        data = s.get_all_version_data(
+            interface='public',
+            service_type='compute')
+        self.assertEqual({
+            'RegionOne': {
+                'public': {
+                    'compute': [{
+                        'collection': None,
+                        'max_microversion': None,
+                        'min_microversion': None,
+                        'next_min_version': None,
+                        'not_before': None,
+                        'raw_status': 'stable',
+                        'status': 'CURRENT',
+                        'url': 'https://compute.example.com/nova/public',
+                        'version': '2.0'
+                    }, {
+                        'collection': None,
+                        'max_microversion': '2.35',
+                        'min_microversion': '2.1',
+                        'next_min_version': None,
+                        'not_before': None,
+                        'raw_status': 'stable',
+                        'status': 'CURRENT',
+                        'url': 'https://compute.example.com/nova/public',
+                        'version': '2.1'
+                    }],
+                }
+            }
+        }, data)
+
+    def test_get_all_version_data_adapter(self):
+        nova_disc = fixture.DiscoveryList(v2=False, v3=False)
+        nova_disc.add_microversion(
+            href=self.TEST_COMPUTE_PUBLIC, id='v2')
+        nova_disc.add_microversion(
+            href=self.TEST_COMPUTE_PUBLIC, id='v2.1',
+            min_version='2.1', max_version='2.35')
+
+        self.stub_url(
+            'GET', [], base_url=self.TEST_COMPUTE_PUBLIC, json=nova_disc)
+        s = session.Session(auth=self.create_auth_plugin())
+        a = adapter.Adapter(session=s, service_type='compute')
+
+        data = a.get_all_version_data()
+        self.assertEqual({
+            'RegionOne': {
+                'public': {
+                    'compute': [{
+                        'collection': None,
+                        'max_microversion': None,
+                        'min_microversion': None,
+                        'next_min_version': None,
+                        'not_before': None,
+                        'raw_status': 'stable',
+                        'status': 'CURRENT',
+                        'url': 'https://compute.example.com/nova/public',
+                        'version': '2.0'
+                    }, {
+                        'collection': None,
+                        'max_microversion': '2.35',
+                        'min_microversion': '2.1',
+                        'next_min_version': None,
+                        'not_before': None,
+                        'raw_status': 'stable',
+                        'status': 'CURRENT',
+                        'url': 'https://compute.example.com/nova/public',
+                        'version': '2.1'
+                    }],
+                }
+            }
+        }, data)
+
+    def test_get_all_version_data_service_alias(self):
+
+        cinder_disc = fixture.DiscoveryList(v2=False, v3=False)
+
+        # The version discovery dict will not have a project_id
+        cinder_disc.add_nova_microversion(
+            href=self.TEST_VOLUME.versions['v3'].discovery.public,
+            id='v3.0', status='CURRENT',
+            min_version='3.0', version='3.20')
+
+        # Adding a v2 version to a service named volumev3 is not
+        # an error. The service itself is cinder and has more than
+        # one major version.
+        cinder_disc.add_nova_microversion(
+            href=self.TEST_VOLUME.versions['v2'].discovery.public,
+            id='v2.0', status='SUPPORTED')
+
+        self.stub_url(
+            'GET', [],
+            base_url=self.TEST_VOLUME.unversioned.public + '/',
+            json=cinder_disc)
+
+        a = self.create_auth_plugin()
+        s = session.Session(auth=a)
+
+        data = s.get_all_version_data(
+            interface='public',
+            service_type='block-store')
+        self.assertEqual({
+            'RegionOne': {
+                'public': {
+                    'block-storage': [{
+                        'collection': None,
+                        'max_microversion': None,
+                        'min_microversion': None,
+                        'next_min_version': None,
+                        'not_before': None,
+                        'raw_status': 'SUPPORTED',
+                        'status': 'SUPPORTED',
+                        'url': 'https://block-storage.example.com/public/v2',
+                        'version': '2.0'
+                    }, {
+                        'collection': None,
+                        'max_microversion': '3.20',
+                        'min_microversion': '3.0',
+                        'next_min_version': None,
+                        'not_before': None,
+                        'raw_status': 'CURRENT',
+                        'status': 'CURRENT',
+                        'url': 'https://block-storage.example.com/public/v3',
+                        'version': '3.0'
+                    }],
+                }
+            }
+        }, data)
+
     def test_endpoint_data_no_version_no_discovery(self):
         a = self.create_auth_plugin()
         s = session.Session(auth=a)
@@ -767,7 +1221,7 @@ class CommonIdentityTests(object):
 
         # We should only try to fetch the versioned discovery url once
         self.requests_mock.get(
-            self.TEST_VOLUME.versions['v3'].discovery.public, resps)
+            self.TEST_VOLUME.versions['v3'].discovery.public + '/', resps)
 
         data = a.get_endpoint_data(session=s,
                                    service_type='volumev3',
@@ -842,7 +1296,7 @@ class CommonIdentityTests(object):
 
         # Fetch v2.0 first - since that doesn't match endpoint optimization,
         # it should fetch the unversioned endpoint
-        v2_data = s.get_endpoint_data(service_type='volumev3',
+        v2_data = s.get_endpoint_data(service_type='block-storage',
                                       interface='public',
                                       min_version='2.0',
                                       max_version='2.latest',
@@ -893,10 +1347,10 @@ class CommonIdentityTests(object):
         self.requests_mock.get(
             self.TEST_VOLUME.unversioned.public + '/', json=disc)
 
-        # We're requesting version 2 of volumev3 to make sure we
+        # We're requesting version 2 of block-storage to make sure we
         # trigger the logic constructing the unversioned endpoint from the
         # versioned endpoint in the catalog
-        s.get_endpoint_data(service_type='volumev3',
+        s.get_endpoint_data(service_type='block-storage',
                             interface='public',
                             min_version='2.0',
                             max_version='2.latest',
@@ -1095,6 +1549,10 @@ class V3(CommonIdentityTests, utils.TestCase):
     def version(self):
         return 'v3'
 
+    @property
+    def discovery_version(self):
+        return '3.0'
+
     def get_auth_data(self, **kwargs):
         kwargs.setdefault('project_id', self.PROJECT_ID)
         token = fixture.V3Token(**kwargs)
@@ -1123,6 +1581,15 @@ class V3(CommonIdentityTests, utils.TestCase):
             internal=self.TEST_VOLUME.versions['v3'].service.internal,
             region=region)
 
+        # Add block-storage as a versioned endpoint so that we can test
+        # versioned to unversioned inference.
+        svc = token.add_service('block-storage')
+        svc.add_standard_endpoints(
+            admin=self.TEST_VOLUME.versions['v3'].service.admin,
+            public=self.TEST_VOLUME.versions['v3'].service.public,
+            internal=self.TEST_VOLUME.versions['v3'].service.internal,
+            region=region)
+
         svc = token.add_service('baremetal')
         svc.add_standard_endpoints(
             internal=self.TEST_BAREMETAL_INTERNAL,
@@ -1150,6 +1617,10 @@ class V2(CommonIdentityTests, utils.TestCase):
     def version(self):
         return 'v2.0'
 
+    @property
+    def discovery_version(self):
+        return '2.0'
+
     def create_auth_plugin(self, **kwargs):
         kwargs.setdefault('auth_url', self.TEST_URL)
         kwargs.setdefault('username', self.TEST_USER)
@@ -1173,7 +1644,8 @@ class V2(CommonIdentityTests, utils.TestCase):
         region = 'RegionOne'
 
         svc = token.add_service('identity')
-        svc.add_endpoint(self.TEST_ADMIN_URL, region=region)
+        svc.add_endpoint(admin=self.TEST_ADMIN_URL, region=region,
+                         public=None, internal=None)
 
         svc = token.add_service('compute')
         svc.add_endpoint(public=self.TEST_COMPUTE_PUBLIC,
@@ -1195,6 +1667,15 @@ class V2(CommonIdentityTests, utils.TestCase):
             internal=self.TEST_VOLUME.versions['v3'].service.internal,
             region=region)
 
+        # Add block-storage as a versioned endpoint so that we can test
+        # versioned to unversioned inferance.
+        svc = token.add_service('block-storage')
+        svc.add_endpoint(
+            admin=self.TEST_VOLUME.versions['v3'].service.admin,
+            public=self.TEST_VOLUME.versions['v3'].service.public,
+            internal=self.TEST_VOLUME.versions['v3'].service.internal,
+            region=region)
+
         svc = token.add_service('baremetal')
         svc.add_endpoint(
             public=None, admin=None,
@@ -1398,6 +1879,8 @@ class CatalogHackTests(utils.TestCase):
                                       max_version=(2, discover.LATEST))
 
         self.assertTrue(common_m.called)
+        self.assertEqual(self.OTHER_URL, data.service_url)
+        self.assertEqual(self.OTHER_URL, data.catalog_url)
         self.assertEqual(self.OTHER_URL, data.url)
         self.assertEqual((2, 1), data.min_microversion)
         self.assertEqual((2, 35), data.max_microversion)
@@ -1928,3 +2411,27 @@ class GenericAuthPluginTests(utils.TestCase):
                               endpoint_filter=self.ENDPOINT_FILTER)
 
         self.assertIn(name, str(e))
+
+
+class DiscoveryFailures(utils.TestCase):
+    TEST_ROOT_URL = 'http://127.0.0.1:5000/'
+
+    def test_connection_error(self):
+        self.requests_mock.get(self.TEST_ROOT_URL,
+                               exc=exceptions.ConnectionError)
+        sess = session.Session()
+        p = identity.generic.password.Password(self.TEST_ROOT_URL)
+        self.assertRaises(exceptions.DiscoveryFailure, p.get_auth_ref, sess)
+
+    def test_client_exception(self):
+        self.requests_mock.get(self.TEST_ROOT_URL,
+                               exc=exceptions.ClientException)
+        sess = session.Session()
+        p = identity.generic.password.Password(self.TEST_ROOT_URL)
+        self.assertRaises(exceptions.ClientException, p.get_auth_ref, sess)
+
+    def test_ssl_error(self):
+        self.requests_mock.get(self.TEST_ROOT_URL, exc=exceptions.SSLError)
+        sess = session.Session()
+        p = identity.generic.password.Password(self.TEST_ROOT_URL)
+        self.assertRaises(exceptions.DiscoveryFailure, p.get_auth_ref, sess)
